/*
OpenMorpho Library - Mathematical Morphology (Image processing) Library
Copyright (C) 2000  Laurent Najman

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can contact the author at
        mailto:laurent@najman.org
and download the latest version at
        http://www.laurentnajman.org
*/

// Tree.cpp: implementation of the CTree class.
//
//////////////////////////////////////////////////////////////////////

#include "Tree.h"
#include <iostream>
#include <math.h>
#include <assert.h>

//#define DEBUG

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTree::CTree(CNode *n)
{
  root = n;
  root->father = NULL;

  nbNodes = 1;
  Euler = NULL;
  Represent = NULL;
  Minim = NULL;
  nbRepresent = 0;
}

CTree::~CTree()
{
  clearMemory();
}


////////////////////////////////
void CTree::addSon(CNode *father, CNode *n) 
{
  if (n == root) {
    root = father; 
    root->father = NULL;
  }

  father->sons.push_back(n);
  n->father = father;
  nbNodes++;
}

////////////////////////////////
void CTree::deleteNode(CNode *node) 
{
  if (node == root) {
    if (node->sons.size() !=1) {
      return; // Can't delete root;
    } else {
      root = (*node->sons.begin());
      root->father = NULL;
      nbNodes--;
      return;
    }
  } 

  CNode * father = node->father;
  ListOfNodesT::iterator it;
  for (it=node->sons.begin(); it!=node->sons.end(); ++it) {
      addSon(father, (*it));
      nbNodes--;
  }
  father->sons.remove(node);
  nbNodes--;
}

////////////////////////////////
void CTree::swap(CNode *father, CNode *son)
{
  if (son->father != father)
    return;
  father->sons.remove(son);
  son->sons.push_back(father);

  if (father == root) {
    root = son;
    root->father=NULL;
  } else {
    CNode *n;
    n = father->father;
    n->sons.remove(father);
    n->sons.push_back(son);
  }
}


////////////////////////////////
void CTree::clearMemory(CNode *tr) 
{
  ListOfNodesT::iterator it;
  
  if (tr == NULL)
    return;
  for (it=tr->sons.begin(); it!=tr->sons.end(); ++it) {
    clearMemory((*it));
  }
  delete (tr);
  tr = NULL;
}

void CTree::clearMemory()
{
  int i;

  clearMemory(root);
  if (Euler != NULL)
    delete [] Euler;
  if (Represent != NULL)
    delete [] Represent;
  if (Minim != NULL) {

    int logn = int(ceil(log(double(nbRepresent))/log(2.0)));

    for (i=0; i< logn; i++)
      delete Minim[i];
    delete [] Minim;
  }
/*
*/
}

////////////////////////////////
void CTree::printTree(CNode *tr, int depth)
{
  ListOfNodesT::iterator it;
  
  if (tr == NULL)
    return;
  for (int i=0; i<depth; i++)
    std::cout << " ";
  std::cout << tr->data << "(" << tr->sons.size() << ")" << std::endl << std::flush;
  for (it=tr->sons.begin(); it!=tr->sons.end(); ++it) {
    printTree((*it), depth+1);
  }
  
}

void CTree::printTree()
{
  printTree(root);
}

void CTree::printTree(FILE *fp, CNode *tr, int depth)
{
  ListOfNodesT::iterator it;
  
  if (tr == NULL)
    return;
  for (int i=0; i<depth; i++)
    fprintf(fp, "    ");
  fprintf(fp, "%4d (%d)\n", tr->data, tr->sons.size());
  for (it=tr->sons.begin(); it!=tr->sons.end(); ++it) {
    printTree(fp, (*it), depth+1);
  }
  
}

void CTree::printTree(FILE *fp)
{
  printTree(fp, root);
}

void CTree::drawTree(FILE *fp, CNode *tr, int depth)
{
  ListOfNodesT::iterator it;
  
  if (tr == NULL)
    return;
  for (int i=0; i<depth; i++)
    fprintf(fp, "    ");
  if (tr->sons.size() != 0)
    fprintf(fp, "%8d [label=\"%d\"];\n", tr, tr->data % 256);
  else
    fprintf(fp, "%8d [shape=polygon,sides=5,label=\"%d\"];\n", tr, tr->data % 256);
  if (tr->father != NULL)
    fprintf(fp, "%8d -> %8d;\n", tr->father, tr);
  for (it=tr->sons.begin(); it!=tr->sons.end(); ++it) {
    drawTree(fp, (*it), depth+1);
  }
  
}

void CTree::drawTree(FILE *fp)
{
  fprintf(fp, "digraph G {\n	size =\"8,10\";\n");
  drawTree(fp, root);
  fprintf(fp, "}\n");
}

/////////////////////////////////////////
// lca : Nearest (Lowest) Common Ancestor
// 
// From: The LCA Problem Revisited 
// M.A. Bender - M. Farach-Colton
//

// Depth-first preprocessing
int CTree::LCApreprocessDepthFirst(CNode *node, int depth)
{
  //static int nbr=-1;  
  //static int rep = 0;
  ListOfNodesT::iterator it;
  
  nbr++;
  Euler[nbr] = node;
  node->Represent = nbr;
#ifdef DEBUG
  cout << "Represent " << node->data << " = " << nbr << "\n";
#endif
  node->depth = depth;
  node->number = rep;
  Represent[rep++] = nbr;
  for (it = node->sons.begin(); it != node->sons.end(); ++it) {
    LCApreprocessDepthFirst((*it), depth+1);
    Euler[++nbr] = node;
#ifdef DEBUG
    cout << "Euler[" << node->data << "] = " << nbr << "\n";
#endif
  }
  return nbr;
}

/*
void CTree::LCApreprocess()
{
  //O(n.log(n)) preprocessing
  
  nbr = -1; // Initialization number of euler nodes
  rep = 0;

  Euler = new CNode* [2*nbNodes-1];
  Represent = new int[nbNodes];

  int nbr = LCApreprocessDepthFirst(root);

  // Check that the number of nodes in the tree was correct
  assert((nbr+1) == (2*nbNodes-1));

  nbRepresent = 2*nbNodes-1;
  int logn = int(ceil(log(double(nbRepresent))/log(2.0)));

  int i,j,k1,k2;
  Minim = new int* [logn];
  Minim[0] = new int[nbRepresent];
  for (i=0; i<nbRepresent; i++) {
    Minim[0][i] = i;
  }

  for (j=1; j<logn; j++) {
    k1 = 1<<(j);
    k2 = 1<<(j-1);
    Minim[j] = new int[nbRepresent - k1 + 1];
#ifdef DEBUG
    cout << "\nj = " << j << "; k1 = " << k1 << "; k2 = " << k2 << "\n";
#endif
    for (i=0; i<nbRepresent - k1 + 1; i++) {
#ifdef DEBUG
      cout << " depth(" << j-1 << "," << i << ") = " << Euler[Minim[j-1][i]]->depth << " - ";
      cout << " depth(" << j-1 << "," << i+k2 << ") = " << Euler[Minim[j-1][i+k2]]->depth << " - ";
#endif
      if ((i+ (k2 << 1)) >= nbRepresent) {
	Minim[j][i] = 0; nbRepresent;
      } else
      if (Euler[Minim[j-1][i]]->depth <= Euler[Minim[j-1][i+k2]]->depth) {
	Minim[j][i] = Minim[j-1][i];
      } else {
	Minim[j][i] = Minim[j-1][i+k2];
      }
#ifdef DEBUG
      cout << "Minim[" << j << "," << i << "] = " << Minim[j][i] <<  "\n";
#endif
    }
  }

}
*/

void CTree::LCApreprocess()
{
  //O(n.log(n)) preprocessing
  
  nbr = -1; // Initialization number of euler nodes
  rep = 0;

  Euler = new CNode* [2*nbNodes-1];
  Represent = new int[nbNodes];

  int nbr = LCApreprocessDepthFirst(root);

  // Check that the number of nodes in the tree was correct
  assert((nbr+1) == (2*nbNodes-1));

  nbRepresent = 2*nbNodes-1;
  int logn = int(ceil(log(double(nbRepresent))/log(2.0)));

  int i,j,k1,k2;
  Minim = new int* [logn];
  Minim[0] = new int[nbRepresent];
  for (i=0; i<nbRepresent-1; i++) {
    if (Euler[i]->depth < Euler[i+1]->depth) {
      Minim[0][i] = i;
    } else {
      Minim[0][i] = i+1;
    }
#ifdef DEBUG
      cout << "Minim[" << 0 << "," << i << "] = " << Minim[0][i] <<  "\n";
#endif
  }
  Minim[0][nbRepresent-1] = nbRepresent-1;
#ifdef DEBUG
      cout << "Minim[" << 0 << "," << nbRepresent-1 << "] = " << Minim[0][nbRepresent-1] <<  "\n";
#endif

  for (j=1; j<logn; j++) {
    k1 = 1<<(j-1);
    k2 = k1<<1;
    Minim[j] = new int[nbRepresent];
#ifdef DEBUG
    cout << "\nj = " << j << "; k1 = " << k1 << "; k2 = " << k2 << "\n";
#endif
    for (i=0; i<nbRepresent; i++) {
#ifdef DEBUG
      //cout << " depth(" << j-1 << "," << i << ") = " << Euler[Minim[j-1][i]]->depth << " - ";
      //cout << " depth(" << j-1 << "," << i+k1 << ") = " << Euler[Minim[j-1][i+k1]]->depth << " - ";
#endif
      if ((i+ k2) >= nbRepresent) {
	Minim[j][i] = nbRepresent-1;
      } else {
	if (Euler[Minim[j-1][i]]->depth <= Euler[Minim[j-1][i+k1]]->depth) {
	  Minim[j][i] = Minim[j-1][i];
	} else {
	  Minim[j][i] = Minim[j-1][i+k1];
	}
      }
#ifdef DEBUG
      cout << "Minim[" << j << "," << i << "] = " << Minim[j][i] <<  "\n";
#endif
    }
  }

}

/*
CNode* CTree::LCAGet(CNode *n1, CNode *n2)
{
  int ii, jj, kk, k;
  ii = n1->Represent;
  jj = n2->Represent;
  if (ii == jj)
    return n1;

  if (ii > jj) {
    kk = jj;
    jj = ii;
    ii = kk;
  }
  k = int(log(double(jj - ii))/log(2.));
#ifdef DEBUG
      cout << " depth(" << k << "," << ii << ") = " << Euler[Minim[k][ii]]->depth << " - ";
      cout << " depth(" << k << "," << jj-(1<<(k))+1 << ") = " << Euler[Minim[k][jj-(1<<(k))+1]]->depth << " - ";
#endif
  if (Euler[Minim[k][ii]]->depth < Euler[Minim[k][jj-(1<<(k))+1]]->depth) {
    return Euler[Minim[k][ii]];
  } else {
    return Euler[Minim[k][jj-(1<<k)+1]];
  }
}
*/

CNode* CTree::LCAGet(CNode *n1, CNode *n2)
{
  int ii, jj, kk, k;
  ii = n1->Represent;
  jj = n2->Represent;
#ifdef DEBUG
  //cout << "Testing node " << ii << " and node " << jj << "\n";
#endif
  if (ii == jj)
    return n1;

  if (ii > jj) {
    kk = jj;
    jj = ii;
    ii = kk;
  }
  k = int(log(double(jj - ii))/log(2.));
#ifdef DEBUG
      cout << " depth(" << k << "," << ii << ") = " << Euler[Minim[k][ii]]->depth << " - ";
      cout << " depth(" << k << "," << jj-(1<<(k)) << ") = " << Euler[Minim[k][jj-(1<<(k))]]->depth << " - ";
#endif
  if (Euler[Minim[k][ii]]->depth < Euler[Minim[k][jj-(1<<(k))]]->depth) {
    return Euler[Minim[k][ii]];
  } else {
    return Euler[Minim[k][jj-(1<<k)]];
  }
}

CNode* CTree::LCAGetSlow(CNode *n1, CNode *n2)
{
  CNode *x;
  x=n1;
  while (x!=NULL) {
    x->number = 0;
    x = x->father;
  }
  x=n2;
  while (x!=NULL) {
    x->number = 0;
    x = x->father;
  }
  x = n1;
  while (x!=NULL) {
    x->number++;
    x = x->father;
  }
  x = n2;
  while (x!=NULL) {
    x->number++;
    x = x->father;
  }
  x = n1;
  while (x->number != 2)
    x = x->father;
  
  return x;
}

int CTree::LCAGet(int r1, int r2)
{
  return LCAGet(Euler[Represent[r1]], Euler[Represent[r2]])->Represent;
}


int CTree::check()
{
  int nbErrors=0;
  for (int i; i <NbNodesGet(); i++) {
    for (int j=i; j< NbNodesGet(); j++) {
      CNode *xx, *yy;
      xx = NodesGet(i);
      yy = NodesGet(j);
      if (LCAGet(xx,yy) != LCAGetSlow(xx,yy)) {
	nbErrors++;
#ifdef DEBUG
	cout << "Error node " << i << "," << j << "\n";
#endif
      }
    }
  }
  return nbErrors;
}

int CTree::testing(int depth, int n, CNode *father)
{
  CNode *node;
  if (depth == 0)
    return -1;
  if (father == NULL) {
    node = new CNode(0);
    root = node;
    root->father = NULL;
    nbNodes = 1;
    Euler = NULL;
    Represent = NULL;
    Minim = NULL;
    nbRepresent = 0;
    //cout <<"father\n";
    testing(depth-1, n, node);
  } else {
    int nbSons = int(double(rand())*double(n)/double(RAND_MAX));
    //cout << nbSons <<"\n";
    for (int i=0; i<nbSons; i++) {
      node = new CNode(nbNodes);
      addSon(father, node);
      //cout << nbNodes << " node\n";
      testing(depth-1, n, node);
    }
  }
  LCApreprocess();
  return check();
}


