/*! \file graph_algos.c
    \brief algorithmes fondamentaux
*/
/*
      Michel Couprie, septembre 1999
*/
#include "graphaux.h"
#include "graphes.h"

/* ====================================================================== */
/*! \fn boolean * CompFortConnexe(graphe * g, graphe *g_1, int a)
    \param g (entre) : un graphe.
    \param g_1 (entre) : le graphe symtrique de g.
    \param a (entre) : un sommet du graphe g.
    \return un sous-ensemble des sommets de g (tableau de boolens).
    \brief retourne la composante fortement connexe de g contenant a 
          (sous la forme d'un tableau de boolens). 
*/
boolean * CompFortConnexe(graphe * g, graphe *g_1, int a)
/* ====================================================================== */
{
  boolean * Ca;  /* pour le resultat */
  boolean * D;   /* pour les "descendants" (successeurs a N niveaux) */
  boolean * A;   /* pour les "ascendants" (predecesseurs a N niveaux) */
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, n, s;
  pcell p;

  n = g->nsom;
  T = CreeLifoVide(n);
  A = EnsembleVide(n);
  D = EnsembleVide(n);
  Ca = EnsembleVide(n);

  /* calcule les descendants de a */
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout sommet s successeur de i */
      s = p->som;
      if (!D[s]) 
      {
        D[s] = TRUE;
        LifoPush(T, s);
      }
    }
  }

  /* calcule les ascendants de a */
  LifoPush(T, a);
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g_1->gamma[i]; p != NULL; p = p->next) 
    { /* pour tout s successeur de i */
      s = p->som;
      if (!A[s]) 
      {
        A[s] = TRUE;
        LifoPush(T, s);
      }
    }
  }

  /* intersection de D et de A, union { a } */
  for (i = 0; i < n; i++) Ca[i] = (D[i] && A[i]);
  Ca[a] = TRUE;
    
  free(A);
  free(D);
  LifoTermine(T);
  return Ca;
} /* CompFortConnexe() */

/* ====================================================================== */
/*! \fn boolean * CompConnexe(graphe * g, graphe *g_1, int a)
    \param   g (entre) : un graphe.
    \param g_1 (entre) : le graphe symtrique de g.
    \param a (entre) : un sommet du graphe g.
    \return un sous-ensemble de sommets de g (tableau de boolens).
    \brief retourne la composante connexe de g contenant a
          (sous la forme d'un tableau de boolens). 
*/
boolean * CompConnexe(graphe * g, graphe *g_1, int a)
/* ====================================================================== */
{
  boolean * Ca;  /* pour le resultat */
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, n, s;
  pcell p;

  n = g->nsom;
  T = CreeLifoVide(n);
  Ca = EnsembleVide(n);

  Ca[a] = TRUE;
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (!Ca[s]) 
      {
        Ca[s] = TRUE;
        LifoPush(T, s);
      }
    }
    for (p = g_1->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i dans g_1 */
      s = p->som;
      if (!Ca[s]) 
      {
        Ca[s] = TRUE;
        LifoPush(T, s);
      }
    }
  }

  LifoTermine(T);
  return Ca;
} /* CompConnexe() */

/* ====================================================================== */
/*! \fn int Connexe(graphe * g, graphe *g_1)
    \param g (entre) : un graphe.
    \param g_1 (entre) : le graphe symtrique de g.
    \return boolen.
    \brief retourne 1 si le graphe est connexe, 0 sinon. 
*/
boolean Connexe(graphe * g, graphe *g_1)
/* ====================================================================== */
{
  int i;
  boolean * C  = CompConnexe(g, g_1, 0);
  boolean ret = TRUE;

  int n = g->nsom;
  for (i = 0; i < n; i++) if (!C[i]) { ret = FALSE; break; }
  free(C);
  return ret;
} /* Connexe() */

/* ====================================================================== */
/* ====================================================================== */
/* ARBRE DE POIDS MAXIMUM */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
/*! \fn graphe * ArbrePoidsMax(graphe * g, graphe *g_1)
    \param g (entre) : un graphe pondr connexe sans boucle.
    \param g_1 (entre) : le graphe symtrique de g.
    \return un arbre. 
    \brief retourne un arbre de poids maximal pour g.
    \warning aucune vrification des donnes initiale n'est effectue,
    \warning les donnes du graphe d'origine seront dtruites. 
*/
graphe * ArbrePoidsMax(graphe * g, graphe *g_1)
/* ====================================================================== */
{
  int n = g->nsom;
  int msav, m = g->narc;
  graphe * apm;    /* pour le resultat */
  int *A;          /* tableau pour ranger les index des arcs */
  int i, j, t, q;
  msav = m;

  /* tri des arcs par poids croissant */
  A = (int *)malloc(m * sizeof(int)); /* allocation index */
  if (A == NULL)
  {   fprintf(stderr, "ArbrePoidsMax : malloc failed\n");
      exit(0);
  }  
  for (i = 0; i < m; i++) A[i] = i; /* indexation initiale */
  TriRapideStochastique(A, g->v_arcs, 0, m-1);

  /* Boucle sur les arcs pris par ordre croissant:
       on enleve un arc, si cela deconnecte le graphe on le remet.
     Arret lorsque m = n-1. */
  i = 0;
  while (m > n-1)
  {
    t = g->tete[A[i]]; q = g->queue[A[i]]; 
    RetireArc(g, t, q);
    RetireArc(g_1, q, t);
    if (Connexe(g, g_1)) 
      m--; 
    else 
    {
      AjouteArc(g, t, q);
      AjouteArc(g_1, q, t);
    }
    i++;
  }

  /* construit le graphe resultat */
  apm = InitGraphe(n, m);
  for (i = 0, j = 0; i < msav; i++)
  {
    t = g->tete[i]; q = g->queue[i]; 
    if (EstSuccesseur(g, t, q)) 
    {
      AjouteArc(apm, t, q);
      apm->tete[j] = t; 
      apm->queue[j] = q; 
      apm->v_arcs[j] = g->v_arcs[i]; 
      j++;
    }
  }

  if (g->x)
    for (i = 0; i < n; i++) { apm->x[i] = g->x[i];  apm->y[i] = g->y[i]; }

  free(A);
  return apm;
} /* ArbrePoidsMax() */

/* ====================================================================== */
/* ====================================================================== */
/* PROGRAMME DE TEST */
/* ====================================================================== */
/* ====================================================================== */

#ifdef TEST
void main(int argc, char **argv)
{
  graphe * g, *g_1, *a;
  int s1, s2, na, ns;
  boolean *Cs;

  if (argc != 3)
  {
    fprintf(stderr, "usage: %s <nombre sommets> <nombre arcs>\n", argv[0]);
    exit(0);
  }

  ns = atoi(argv[1]);
  na = atoi(argv[2]);
  
  g = GrapheAleatoire(ns, na);
/*
  AfficheSuccesseurs(g);
  AfficheArcs(g);
*/
  g_1 = Symetrique(g);
/*
  AfficheSuccesseurs(g_1);
  AfficheArcs(g_1);
*/

  if (Connexe(g, g_1))
  {
    printf("graphe connexe\n");
    a = ArbrePoidsMax(g, g_1);
/*
    AfficheSuccesseurs(a);
    AfficheArcs(a);
*/
    TermineGraphe(a);
  }
  else printf("graphe NON connexe\n");

/*
  do
  {
    printf("entrer un sommet : ");
    scanf("%d", &s1);
    if (s1 >= 0) 
    {
      Cs = CompFortConnexe(g, g_1, s1);
      AfficheEnsemble(Cs, g->nsom);
      free(Cs);
    }
  } while (s1 >= 0);

  do
  {
    printf("entrer un sommet : ");
    scanf("%d", &s1);
    if (s1 >= 0) 
    {
      Cs = CompConnexe(g, g_1, s1);
      AfficheEnsemble(Cs, g->nsom);
      free(Cs);
    }
  } while (s1 >= 0);
*/

  TermineGraphe(g);
  TermineGraphe(g_1);

} /* main() */
#endif
