/*! \file graphes.h
    \brief structures de base pour la manipulation de graphes
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#define M_PI	 3.14159265358979323846
#define max(X,Y) ((X)>=(Y)?(X):(Y))
#define min(X,Y) ((X)<=(Y)?(X):(Y))

/* ================================================ */
/* types publics */
/* ================================================ */

/*! \struct cell
    \brief structure de cellule pour les listes chanees de successeurs.
*/
typedef struct cell {
//! index du sommet
  int som;
//! poids de l'arc pour les graphes pondrs 
  double v_arc; 
//! suite de la liste ou pointeur NULL
  struct cell * next; 
} cell;

/*! \var pcell
    \brief pointeur sur une cellule
*/
typedef cell * pcell; 


/*! \struct graphe
    \brief structure pour la representation des graphes
*/
typedef struct graphe {

  /* informations globales */

//!  nombre de sommets 
  int nsom;         
//!  nombre maximum d'arcs
  int nmaxarc;      
//!  nombre d'arcs
  int narc;         

  /* representation par listes chainees de successeurs (application gamma) */

//!  tableau des cellules en rserve 
  pcell reserve;    
//!  liste des cellules libres gre en pile lifo 
  pcell libre;      
//!  tableau des listes de successeurs index par les sommets 
  pcell * gamma;    

  /* representation par liste d'arcs 
     (vecteurs tete (sommet initial), queue (sommet final)) */

//!  tableau des extremits initiales d'arcs 
  int *tete;        
//!  tableau des extremits finales d'arcs 
  int *queue;       

  /* informations additionelles ajoutees aux arcs */

//!  tableau des valeurs associes aux arcs
  double *v_arcs;

  /* informations additionelles ajoutees aux sommets */

//!  tableau des valeurs associes aux sommets
  double *v_sommets;

//!  abcisses des sommets 
  double *x;        
//!  ordonnees des sommets 
  double *y;        
//!  noms des sommets 
  char **nomsommet; 
} graphe;

/* ================================================ */
/* prototypes */
/* ================================================ */

/* ====================================================================== */
void AfficheEnsemble(boolean *s, int n);
/* Donnee s : un tableau de valeurs booleennes.
   Donnee n : la taille du tableau.
   Fonction : affiche l'ensemble represente par le tableau de booleens s. */
/* ====================================================================== */

/* ====================================================================== */
void AfficheListe(pcell p);
/* Donnee p : une liste chainee de successeurs.
   Fonction : affiche le contenu de la liste p. */
/* ====================================================================== */

/* ====================================================================== */
void AfficheSuccesseurs(graphe * g) ;
/* Donnee g : un graphe.
   Fonction : affiche le graphe dans sa representation "listes de successeurs". */
/* ====================================================================== */

/* ====================================================================== */
void AfficheArcs(graphe * g);
/* Donnee g : un graphe.
   Fonction : affiche le graphe dans sa representation "listes d'arcs". */
/* ====================================================================== */

/* ====================================================================== */
void AfficheValeursSommets(graphe * g);
/* ====================================================================== */

/* ====================================================================== */
void PSGraphe(graphe * g, char *filename, double r, double t, double marge);
/* Donnee g : un graphe.
   Donnee filename : nom du fichier postscript a generer.
   Donnee r : rayon des cercles qui representent les sommets (0 pour ne pas les dessiner).
   Donnee t : taille (demi-longueur) des fleches pour les arcs (0 pour ne pas les dessiner).
   Donnee marge : marge en x et en y.
   Fonction : genere une figure PostScript d'apres la representation "successeurs"
              du graphe g. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* FONCTIONS DE PLONGEMENT DE GRAPHES DANS LE PLAN */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
void AutoNomsSommets(graphe * g, int mode);
/* Donnee g : un graphe.
   Donnee mode : 1 pour que les noms soient les indices des sommets,
                 2 pour que les noms soient les valeurs associes aux sommets,
                 3 pour des noms composes de l'indice et de la valeur.
   Fonction : gnre automatiquement des noms pour les sommets du graphe g.  */
/* ====================================================================== */

/* ====================================================================== */
void PlongementCirculaire(graphe * g, double r);
/* Donnee g : un graphe.
   Donnee r : le rayon du cercle.
   Resultat g : le graphe g modifie.
   Fonction : affecte a chaque sommet des coord. (x,y) regulierement reparties
              sur un cercle. */
/* ====================================================================== */

/* ====================================================================== */
void PlongementRadial(graphe * g, int c);
/* Donnee g : un graphe.
   Donnee c : le sommet a placer au centre.
   Resultat g : le graphe g modifie.
   Fonction : repartit les sommets de g sur des cercles concentriques en fonction 
              de leur rang, avec le sommet c au centre. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* FONCTIONS SUR LES LISTES CHAINEES DE SOMMETS */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
pcell AlloueCell(pcell * plibre);
/* Donnee plibre : pointeur sur une liste chainee de cellules libres.
   Resultat : pointeur sur une cellule.
   Fonction : retire la premiere cellule de la liste pointee par plibre
              et retourne un pointeur sur cette cellule */
/* ====================================================================== */

/* ====================================================================== */
void LibereCell(pcell * plibre, pcell p);
/* Donnee plibre : pointeur sur une liste chainee de cellules libres.
   Donnee p : pointeur sur une cellule.
   Fonction : insere la cellule p au debut de la liste pointee par plibre */
/* ====================================================================== */

/* ====================================================================== */
void RetireTete(pcell * plibre, pcell * pliste);
/* Donnee plibre : pointeur sur une liste chainee de cellules libres.
   Donnee pliste : pointeur sur une liste.
   Fonction : retire la premiere cellule de la liste 'pliste'. La cellule
              est chainee a la liste plibre. */
/* ====================================================================== */

/* ====================================================================== */
void AjouteTete(pcell * plibre, pcell * pliste, int a);
/* Donnee plibre : pointeur sur une liste chainee de cellules libres.
   Donnee pliste : pointeur sur une liste.
   Donnee a : un index de sommet.
   Fonction : ajoute une cellule contenant la valeur 'a' en tete de la liste
              'pliste'. La cellule est prise dans la liste 'plibre'. */
/* ====================================================================== */

/* ====================================================================== */
int EstDansListe(pcell p, int a);
/* Donnee p : une liste chainee de successeurs.
   Donnee a : une valeur entiere.
   Resultat : booleen.
   Fonction : retourne 1 si la valeur 'a' se trouve dans la liste p, 0 sinon. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* FONCTIONS D'ALLOCATION / LIBERATION POUR UN GRAPHE */
/* ====================================================================== */
/* ====================================================================== */


/* ====================================================================== */
graphe * InitGraphe(int nsom, int nmaxarc);
/* Donnee nsom : nombre de sommets.
   Donnee nmaxarc : nombre maximum d'arcs.
   Resultat : un graphe.
   Fonction : alloue la memoire necessaire pour representer un graphe a 'nsom' sommets,
              possedant au maximum 'nmaxarc' arcs. 
              Retourne un pointeur sur la structure allouee. */
/* ====================================================================== */

/* ====================================================================== */
void TermineGraphe(graphe * g);
/* Donnee g : un graphe.
   Fonction : libere la memoire occupee par le graphe g. */
/* ====================================================================== */

/* ====================================================================== */
graphe * ReadGraphe(char * filename);
/* Donnee filename : nom du fichier graphe.
   Resultat : un graphe.
   Fonction : Lit les donnees d'un graphe dans le fichier filename,
              retourne un pointeur sur la structure graphe construite. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* FONCTIONS DE MANIPULATION DES ARCS (APPLICATION GAMMA UNIQUEMENT) */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
void AjouteArc(graphe * g, int i, int s);
/* Donnee g : un graphe.
   Donnee i, s : deux sommets de g.
   Resultat g : le graphe modifie.
   Fonction : ajoute l'arc (i,s) au graphe g (application gamma seulement). */
/* ====================================================================== */

/* ====================================================================== */
void RetireArc(graphe * g, int i, int s);
/* Donnee g : un graphe.
   Donnee i, s : deux sommets de g.
   Resultat g : le graphe modifie.
   Fonction : retire l'arc (i,s) du graphe g (application gamma seulement),
              si celui-ci etait present. Sinon, pas d'action. */
/* ====================================================================== */

/* ====================================================================== */
int EstSuccesseur(graphe *g, int i, int s);
/* Donnee g : un graphe.
   Donnee i, s : deux sommets de g.
   Resultat : booleen.
   Fonction : retourne 1 si le sommet 's' est un successeur du sommet 'i', 0 sinon  */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* FONCTIONS DE GENERATION DE GRAPHES */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
graphe * GrapheAleatoire(int nsom, int narc);
/* Donnee nsom : nombre de sommets.
   Donnee narc : nombre d'arcs.
   Resultat : un graphe.
   Fonction : retourne un graphe aleatoire a 'nsom' sommets et 'narc' arcs. 
              Le graphe est asymetrique et sans boucle.
              Le nombre d'arcs doit etre <= nsom (nsom - 1) / 2.
   Remarque : la methode employee ici est naive, tres inefficace du point de vue
              temps de calcul et ne garantit aucune propriete statistique. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* OPERATEURS DE BASE SUR LES GRAPHES */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
graphe * Sym(graphe * g);
/* Donnee g : un graphe.
   Resultat : un graphe.
   Fonction : construit et retourne le graphe g_1 symetrique du graphe g. 
   ATTENTION : seule la representation 'gamma' est utilisee. */
/* ====================================================================== */

/* ====================================================================== */
boolean * CompFortConnexe(graphe * g, graphe *g_1, int a);
/* Donnee g : un graphe.
   Donnee g_1 : le graphe symetrique de g.
   Donnee a : un sommet du graphe g.
   Resultat : un sous-ensemble des sommets de g (tableau de booleens).
   Fonction : retourne la composante fortement connexe de g contenant a. */
/* ====================================================================== */

/* ====================================================================== */
boolean * CompConnexe(graphe * g, graphe *g_1, int a);
/* Donnee g : un graphe.
   Donnee g_1 : le graphe symetrique de g.
   Donnee a : un sommet du graphe g.
   Resultat : un sous-ensemble de sommets de g (tableau de booleens).
   Fonction : retourne la composante connexe de g contenant a. */
/* ====================================================================== */

/* ====================================================================== */
boolean Connexe(graphe * g, graphe *g_1);
/* Donnee g : un graphe.
   Donnee g_1 : le graphe symetrique de g.
   Resultat : booleen.
   Fonction : retourne TRUE si le graphe est connexe, FALSE sinon. */
/* ====================================================================== */

/* ====================================================================== */
/* ====================================================================== */
/* ARBRE DE POIDS MAXIMUM */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
graphe * ArbrePoidsMax(graphe * g, graphe *g_1);
/* Donnee g : un graphe pondere connexe sans boucle.
   Donnee g_1 : le graphe symetrique de g.
   Resultat : un arbre. 
   Fonction : retourne un arbre de poids maximal pour g.
   ATTENTION : - aucune verification des donnees initiale n'est effectuee,
               - les donnees du graphe d'origine seront detruites. */
/* ====================================================================== */
