#include <stdlib.h>
#include <stdio.h>
#include "graphaux.h"
#include "graphes.h"

#define USAGE "verifie si le graphe est biparti, et exhibe un cycle impair dans le cas contraire"

/* ====================================================================== */
graphe * symetrique(graphe * g)
/* ====================================================================== */
{
  graphe *g_1;
  int nsom, narc, al_arcs, k, i, j;
  pcell p;

  nsom = g->nsom;
  narc = g->narc;
  g_1 = InitGraphe(nsom, narc);

  for (i = 0; i < nsom; i++) /* pour tout i sommet de g */
  {
    p = g->gamma[i]; /* acces a la liste des successeurs de i */
    for ( ; p != NULL; p = p->next) /* pour tout j successeur de i */
    {
      j = p->som;
      AjouteArc(g_1, j, i);
    }
  }

  return g_1;
} /* symetrique() */

/* ====================================================================== */
/*! \fn boolean Biparti(graphe * g, graphe *g_1)
    \param   g (entre) : un graphe.
    \param g_1 (entre) : le graphe symtrique de g.
    \return TRUE si le graphe g est biparti, FALSE sinon
    \brief vrifie si le graphe g est biparti, si oui on peut rcuprer la 
           bipartition dans g->v_sommets, sous la forme de deux "couleurs" 0 et 1,
           sinon les valeurs dans g->v_sommets indiquent un cycle de longueur impaire.
*/
boolean Biparti(graphe * g, graphe *g_1)
/* ====================================================================== */
{
  Lifo * T1;   /* liste temporaire geree en pile (Last In, First Out) */
  Lifo * T2;   /* liste temporaire geree en pile (Last In, First Out) */
  Lifo * T;    /* pour l'echange */
  int i, n, s, t, k;
  pcell p;
  boolean * C;  /* pour eviter de reparcourir un sommet plusieurs fois */
  int couleur;
  int * back; /* le "fil d'Arianne" (pointeurs arriere) */

  n = g->nsom;
  T1 = CreeLifoVide(n);
  T2 = CreeLifoVide(n);
  C = EnsembleVide(n);
  back = (int *)malloc(n * sizeof(int));

  C[0] = TRUE;
  LifoPush(T1, 0);  /* on part d'un sommet arbitraire (0 en l'occurence) */
  couleur = 0;
  g->v_sommets[0] = couleur;
  while (!LifoVide(T1)) /* boucle jusqu'a stabilite */
  {
    couleur++;
    couleur = couleur % 2;
    while (!LifoVide(T1)) /* boucle pour le parcours d'un niveau */
    {
      i = LifoPop(T1);
      for (p = g->gamma[i]; p != NULL; p = p->next)
      { /* pour tout s successeur de i */
        s = p->som;
        if (!C[s]) 
        {
          C[s] = TRUE;
          LifoPush(T2, s);
          g->v_sommets[s] = couleur;
          back[s] = i;
        }
        else
	{
          if (g->v_sommets[s] != couleur) goto NONBIPARTI;
	}
      }
      for (p = g_1->gamma[i]; p != NULL; p = p->next)
      { /* pour tout s successeur de i */
        s = p->som;
        if (!C[s]) 
        {
          C[s] = TRUE;
          LifoPush(T2, s);
          g->v_sommets[s] = couleur;
          back[s] = i;
        }
        else
	{
          if (g->v_sommets[s] != couleur) goto NONBIPARTI;
	}
      }
    } // fin du niveau
    T = T2;
    T2 = T1;
    T1 = T;
  }

  LifoTermine(T1);
  LifoTermine(T2);
  free(C);
  free(back);
  return TRUE;

NONBIPARTI:
  // re-initialisation de T1 et du champ v_sommets
  LifoFlush(T1);
  for (k = 0; k < n; k++) g->v_sommets[k] = 0;

  // remonte les pointeurs "back" en partant de s jusqu'au sommet 0
  // stocke la valeur -1 dans le champ v_sommets des sommets parcourus (sauf s)
  t = s;
  do
  { t = back[t];
    g->v_sommets[t] = -1;
  } while (t != 0);

  // remonte les pointeurs "back" en partant de i jusqu'a un sommet marque "-1"
  // stocke les sommets dans T1 (y compris le dernier)
  t = i;
  while (g->v_sommets[t] != -1) 
  { LifoPush(T1, t);
    t = back[t];
  } // while (g->v_sommets[t] != -1) 
  LifoPush(T1, t);

  // remonte les pointeurs "back" en partant de s jusqu'au sommet t (le dernier empile)
  // numerote les sommets
  k = 1;
  do
  { g->v_sommets[s] = k;
    k++;
    s = back[s];
  } while (s != t);

  // retire les sommets de la lifo T1 
  // numerote les sommets
  while (!LifoVide(T1))
  {
    s = LifoPop(T1);
    g->v_sommets[s] = k;
    k++;
  } // while (!LifoVide(T1))

  LifoTermine(T1);
  LifoTermine(T2);
  free(C);
  free(back);
  return FALSE;
} /* Biparti() */

/* ====================================================================== */
int main(int argc, char **argv)
/* ====================================================================== */
{
  graphe *g, *g_1;

  if (argc != 2)
  {
    fprintf(stderr, "usage: %s <nom fichier>\n%s\n", argv[0], USAGE);
    exit(0);
  }

  g = ReadGraphe(argv[1]); /* lit le graphe a partir du fichier */
  g_1 = symetrique(g);            /* calcule le symetrique de g */
  if (Biparti(g, g_1))
    printf("graphe biparti\n");
  else
    printf("graphe NON biparti\n");
  AfficheValeursSommets(g);
  PlongementCirculaire(g, 150);        /* plonge le graphe dans le plan */
  AutoNomsSommets(g, 3);               /* noms des sommet = indices + valeurs */
  PSGraphe(g, "tmp.eps", 4, 10, 60);   /* genere une figure en PostScript */

  TermineGraphe(g);
  TermineGraphe(g_1);

  return 0;
} /* main() */
