#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "graphaux.h"
#include "graphes.h"

/* ====================================================================== */
/*! \fn graphe * PathTree(graphe * g, int d, int a)
    \param g (entre) : un graphe (arbre orient de la racine vers les feuilles)
    \param d (entre) : un sommet (dpart)
    \param a (entre) : un sommet (arrive)
    \return un graphe
    \brief retourne le chemin de \b d vers \b a dans \b g , reprsent par un graphe.
    Les sommets n'apparaissant pas dans le chemin reoivent la valeur 0 dans leur 
    champ v_sommets, les autres la valeur 1.
    
*/
graphe * PathTree(graphe * g, int d, int a)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "PathTree"
{
  int n = g->nsom;
  int m = g->narc;
  graphe * pcc = InitGraphe(n, n-1); /* pour le resultat */
  int i, y, x = a;
  graphe * g_1 = Symetrique(g);
  pcell p;

  for (i = 0; i < n; i++) pcc->v_sommets[i] = 0;

  while (x != d)
  {
    p = g_1->gamma[x]; 
    if (p == NULL)
    {
      fprintf(stderr,"%s: destination node not found\n", F_NAME);
      exit(0);
    }

#define CHECK_TREE
#ifdef CHECK_TREE
    if (p->next != NULL)
    {
      fprintf(stderr,"%s: non-tree structure\n", F_NAME);
      exit(0);
    }
#endif
    
    y = p->som;
    AjouteArcValue(pcc, y, x, p->v_arc);
    pcc->v_sommets[x] = 1;
    pcc->v_sommets[y] = 1;
    x = y;
  }
  TermineGraphe(g_1);
  return pcc;
} /* PathTree() */

/* ====================================================================== */
/*! \fn graphe * FindCycle(graphe * g, graphe * t, graphe * t1)
    \param g (entre) : un graphe
    \param t (entre) : un arbre couvrant sur g "prpar": racine = 0, arcs valus, 
    distances (sommes des poids des arcs)  la racine dans v_sommets
    \param t1 (entre) : un arbre couvrant sur g "prpar": racine = 0, 
    distances (nombre d'arcs)  la racine dans v_sommets
*/
graphe * FindCycle(graphe * g, graphe * t, graphe * t1)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "FindCycle"
{
  int i, a, b, c, Pa, Pb, Pc, Pa1, Pb1, Pc1, n = g->nsom, m = g->narc;
  boolean * T = EnsembleVide(m);
  double *L;
  int imaxL;
  graphe *c1, *c2;
  graphe * ti = Symetrique(t);

  L = (double *)malloc(m * sizeof(double));
  if (L == NULL)
  {
    fprintf(stderr, "%s: malloc failed\n", F_NAME);
    exit(0);
  }
  
  for (i = 0; i < m; i++) // met dans T les indices des arcs de l'arbre t
  {
    a = g->tete[i];
    b = g->queue[i];
    if (EstSuccesseur(t, a, b) || EstSuccesseur(t, b, a)) T[i] = TRUE;
  } // for (i = 0; i < m; i++)

  for (i = 0; i < m; i++) 
    if (!T[i]) // parcourt les artes de g non dans t
    {
      a = g->tete[i];
      b = g->queue[i];
      c = LCA(ti, a, b);
      Pa = t->v_sommets[a];
      Pb = t->v_sommets[b];
      Pc = t->v_sommets[c];
      Pa1 = t1->v_sommets[a];
      Pb1 = t1->v_sommets[b];
      Pc1 = t1->v_sommets[c];
      L[i] = (double)Pa + Pb - (2*Pc); // poids du co-chemin associ  l'arte ab
      L[i] = L[i] / (Pa1 + Pb1 - (2*Pc1));
#ifdef VERBOSE
      printf("arete %d,%d : c = %d, Pa = %d, Pb = %d, Pc = %d, L = %g\n", a, b, c, Pa, Pb, Pc, L[i]);
#endif
    } // for (i = 0; i < m; i++)
    else
      L[i] = 0.0;

  imaxL = 0; // cherche l'arte de co-chemin maximal pour t
  for (i = 1; i < m; i++) if (L[i] > L[imaxL]) imaxL = i;

  a = g->tete[imaxL];
  b = g->queue[imaxL];
  c = LCA(ti, a, b);

#ifdef VERBOSE
  printf("arete de co-chemin max :  %d,%d ; LCA : %d\n", a, b, c);
#endif

  c1 = PathTree(t, c, a);
  c2 = PathTree(t, c, b);
  UnionGraphes(c1, c2);
  AjouteArcValue(c1, a, b, g->v_arcs[imaxL]);

  free(T);
  free(L);
  TermineGraphe(ti);
  TermineGraphe(c2);
  return c1;

} /* FindCycle() */

int main(int argc, char **argv)
{
  graphe *g, *g_1, *a, *a_1, *t, *t1, *c;

  if (argc != 3)
  {
    fprintf(stderr, "usage: %s in.graph out.graph\n", argv[0]);
    exit(0);
  }

  g = ReadGraphe(argv[1]);  // lit le graphe a partir du fichier
  Gamma2ListArcs(g); 
  g_1 = Symetrique(g);
  a = Kruskal1(g, g_1);     // arbre de poids maximum
  a_1 = Symetrique(a);
  t = RootTree(a, a_1, 0);  // arborescence
  DistRoot(t, 0);           // distances  la racine
  t1 = CopyGraphe(t);
  UnitLength(t1);           // met les longueurs d'arcs  1
  DistRoot(t1, 0);          // distances  la racine (nombres d'arcs)

  c = FindCycle(g, t, t1);

  CopyXY(g, c);
  SaveGraphe(c, argv[argc - 1]); 

  TermineGraphe(g);
  TermineGraphe(g_1);
  TermineGraphe(a);
  TermineGraphe(a_1);
  TermineGraphe(t);
  TermineGraphe(t1);
  TermineGraphe(c);

  return 0;
} /* main() */
