/*! \file graph_algos.c
    \brief algorithmes fondamentaux
*/
/*
      Michel Couprie, septembre 1999

      janvier 2005 : DepthTree, MaxDiameterTree, Lee
      fvrier 2005 : LCA (non efficace)
*/
#include "graphes.h"
#define L_INFINITY LONG_MAX

/* ====================================================================== */
/* ====================================================================== */
/* OPERATEURS DE BASE SUR LES GRAPHES */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
/*! \fn int Degre(graphe * g, int s)
    \param g (entre) : un graphe
    \param s (entre) : un sommet
    \return un entier
    \brief retourne le degr du sommet s dans le graphe graphe g (nombre de successeurs)
    \warning seule la reprsentation 'gamma' est utilise
*/
int Degre(graphe * g, int s)
/* ====================================================================== */
{
  int d = 0;
  pcell p;
  for (p = g->gamma[s]; p != NULL; p = p->next) d++;
  return d;
} /* Degre() */

/* ====================================================================== */
/* A RETIRER POUR LE TP 1 */
/* ====================================================================== */
/*! \fn graphe * Symetrique(graphe * g)
    \param g (entre) : un graphe
    \return un graphe
    \brief construit et retourne le graphe g_1 symtrique du graphe g (algorithme linaire)
    \warning seule la reprsentation 'gamma' est utilise
*/
graphe * Symetrique(graphe * g)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Symetrique"
{
  graphe *g_1;
  int nsom, narc, al_arcs, k, i, j;
  pcell p;

  nsom = g->nsom;
  narc = g->narc;
  g_1 = InitGraphe(nsom, narc);
  for (i = 0; i < nsom; i++) /* pour tout i sommet de g */
  {
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout j successeur de i */
      j = p->som;
      AjouteArcValue(g_1, j, i, p->v_arc);
    } // for p
  } // for i

  return g_1;
} /* Symetrique() */

/* ====================================================================== */
/* A RETIRER POUR LE TP 1 */
/* ====================================================================== */
/*! \fn graphe * FermetureSymetrique(graphe * g)
    \param g (entre) : un graphe
    \return un graphe
    \brief construit et retourne la fermeture symtrique du graphe g (algorithme linaire)
    \warning seule la reprsentation 'gamma' est utilise
*/
graphe * FermetureSymetrique(graphe * g)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "FermetureSymetrique"
{
  graphe *gs;
  int nsom, narc, al_arcs, k, i, j;
  pcell p;

  nsom = g->nsom;
  narc = g->narc;
  gs = InitGraphe(nsom, 2 * narc);
  for (i = 0; i < nsom; i++) /* pour tout i sommet de g */
  {
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout j successeur de i */
      j = p->som;
      AjouteArcValue(gs, i, j, p->v_arc);
      AjouteArcValue(gs, j, i, p->v_arc);
    } // for p
  } // for i

  return gs;
} /* FermetureSymetrique() */

/* ====================================================================== */
/*! \fn void CompFortConnexe(graphe * g, graphe *g_1, int a, boolean * Ca)
    \param g (entre) : un graphe
    \param g_1 (entre) : le graphe symtrique de g
    \param a (entre) : un sommet du graphe g
    \param Ca (sortie) : un sous-ensemble des sommets de g (tableau de boolens)
    \brief retourne dans Ca la composante fortement connexe de g contenant a 
          (sous la forme d'un tableau de boolens)
    \warning Ca doit avoir t allou correctement (pas de vrification)
*/
void CompFortConnexe(graphe * g, graphe *g_1, int a, boolean * Ca)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "CompFortConnexe"
{
  boolean * D;   /* pour les "descendants" (successeurs a N niveaux) */
  boolean * A;   /* pour les "ascendants" (predecesseurs a N niveaux) */
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, n, s;
  pcell p;

  n = g->nsom;
  T = CreeLifoVide(n);
  A = EnsembleVide(n);
  D = EnsembleVide(n);
  memset(Ca, 0, n);   // Ca := vide

  /* calcule les descendants de a */
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout sommet s successeur de i */
      s = p->som;
      if (!D[s]) 
      {
        D[s] = TRUE;
        LifoPush(T, s);
      }
    } // for p
  } // while (!LifoVide(T))

  /* calcule les ascendants de a */
  LifoPush(T, a);
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g_1->gamma[i]; p != NULL; p = p->next) 
    { /* pour tout s successeur de i */
      s = p->som;
      if (!A[s]) 
      {
        A[s] = TRUE;
        LifoPush(T, s);
      }
    } // for p
  } // while (!LifoVide(T))

  /* intersection de D et de A, union { a } */
  for (i = 0; i < n; i++) Ca[i] = (D[i] && A[i]);
  Ca[a] = TRUE;
    
  free(A);
  free(D);
  LifoTermine(T);
} /* CompFortConnexe() */

/* ====================================================================== */
/*! \fn boolean ExisteCircuit(graphe * g, int a)
    \param g (entre) : un graphe
    \param a (entre) : un sommet du graphe g
    \return boolen
    \brief teste l'existence d'un circuit dans g contenant a 
*/
boolean ExisteCircuit(graphe * g, int a)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "ExisteCircuit"
{
  boolean * D;   /* pour les "descendants" (successeurs a N niveaux) */
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, n, s;
  pcell p;

  n = g->nsom;
  T = CreeLifoVide(n);
  D = EnsembleVide(n);

  /* calcule les descendants de a */
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout sommet s successeur de i */
      s = p->som;
      if (s == a) { free(D); LifoTermine(T); return TRUE; }
      if (!D[s]) 
      {
        D[s] = TRUE;
        LifoPush(T, s);
      }
    } // for p
  } // while (!LifoVide(T))

  free(D);
  LifoTermine(T);
  return FALSE;
} /* ExisteCircuit() */

/* ====================================================================== */
/*! \fn void CompConnexe(graphe * g, graphe *g_1, int a, boolean * Ca)
    \param   g (entre) : un graphe
    \param g_1 (entre) : le graphe symtrique de g
    \param a (entre) : un sommet du graphe g
    \return un sous-ensemble de sommets de g (tableau de boolens)
    \brief retourne la composante connexe de g contenant a
          (sous la forme d'un tableau de boolens)
*/
void CompConnexe(graphe * g, graphe *g_1, int a, boolean * Ca)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "CompConnexe"
{
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, n, s;
  pcell p;

  n = g->nsom;
  T = CreeLifoVide(n);
  memset(Ca, 0, n); // Ca := vide

  Ca[a] = TRUE;
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (!Ca[s]) 
      {
        Ca[s] = TRUE;
        LifoPush(T, s);
      }
    } // for p
    for (p = g_1->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i dans g_1 */
      s = p->som;
      if (!Ca[s]) 
      {
        Ca[s] = TRUE;
        LifoPush(T, s);
      }
    } // for p
  } // while (!LifoVide(T))

  LifoTermine(T);
} /* CompConnexe() */

/* ====================================================================== */
/*! \fn int Connexe(graphe * g, graphe *g_1)
    \param g (entre) : un graphe
    \param g_1 (entre) : le graphe symtrique de g
    \return boolen
    \brief retourne 1 si le graphe est connexe, 0 sinon
*/
boolean Connexe(graphe * g, graphe *g_1)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Connexe"
{
  int i;
  boolean ret = TRUE;
  int n = g->nsom;
  boolean * C = EnsembleVide(n);
  CompConnexe(g, g_1, 0, C);
  for (i = 0; i < n; i++) if (!C[i]) { ret = FALSE; break; }
  free(C);
  return ret;
} /* Connexe() */

/* ====================================================================== */
/*! \fn boolean CircuitNiveaux(graphe * g)
    \param g (entre) : un graphe
    \return boolen
    \brief si le graphe possde au moins un circuit, retourne FALSE ;
    sinon, calcule les niveaux des sommets du graphes (dans v_sommets) et retourne TRUE
*/
boolean CircuitNiveaux(graphe * g, int a)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "CircuitNiveaux"
{
  int * D;   /* pour les degrs intrieurs */
  int i, n, s;
  pcell p;

  n = g->nsom;

  // A FINIR

  free(D);
  return FALSE;
} /* CircuitNiveaux() */

/* ====================================================================== */
/* ====================================================================== */
/* ARBRES */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2 */
/* ====================================================================== */
/*! \fn graphe * Kruskal1(graphe * g, graphe *g_1)
    \param g (entre) : un graphe pondr connexe antisymtrique antirflexif
    \param g_1 (entre) : le graphe symtrique de g
    \return un arbre
    \brief retourne un arbre de poids maximal pour g
*/
graphe * Kruskal1(graphe * g, graphe *g_1)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Kruskal1"
{
  int n = g->nsom;
  int m = g->narc;
  graphe * apm;    /* pour le resultat */
  graphe * apm_1;  /* pour la detection de cycles */
  int *A;          /* tableau pour ranger les index des arcs */
  int i, j, t, q;
  boolean *Ct = EnsembleVide(n);

  /* tri des arcs par poids croissant */
  A = (int *)malloc(m * sizeof(int)); /* allocation index */
  if (A == NULL)
  {   fprintf(stderr, "%s: malloc failed\n", F_NAME);
      exit(0);
  }  
  for (i = 0; i < m; i++) A[i] = i; /* indexation initiale */
  TriRapideStochastique(A, g->v_arcs, 0, m-1);

  /* construit le graphe resultat, initialement vide */
  apm = InitGraphe(n, n-1);
  apm_1 = InitGraphe(n, n-1);

  /* Boucle sur les arcs pris par ordre decroissant:
     on ajoute un arc dans apm, si cela ne cree pas un cycle dans apm.
     Arret lorsque nb arcs ajoutes = n-1. */

  j = 0; 
  i = m - 1;
  while (j < n-1)
  {
    t = g->tete[A[i]]; q = g->queue[A[i]]; 
    CompConnexe(apm, apm_1, t, Ct);
    if (!Ct[q]) 
    {
      AjouteArcValue(apm, t, q, g->v_arcs[A[i]]);
      AjouteArc(apm_1, q, t);
      j++; 
    }
    i--;
    if ((i < 0) && (j < (n-1)))
    {   fprintf(stderr, "%s: graphe d'origine non connexe\n", F_NAME);
        exit(0);
    }  
  } // while (j < n-1)

  if (g->x != NULL) // recopie les coordonnees des sommets pour l'affichage
    for (i = 0; i < n; i++) { apm->x[i] = g->x[i];  apm->y[i] = g->y[i]; }

  free(A);
  free(Ct);
  TermineGraphe(apm_1);
  return apm;
} /* Kruskal1() */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2 */
/* ====================================================================== */
/*! \fn graphe * Kruskal2(graphe * g, graphe *g_1)
    \param g (entre) : un graphe pondr connexe antisymtrique antirflexif
    \param g_1 (entre) : le graphe symtrique de g
    \return un arbre
    \brief retourne un arbre de poids maximal pour g
    \warning les donnes du graphe d'origine seront dtruites
*/
graphe * Kruskal2(graphe * g, graphe *g_1)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Kruskal2"
{
  int n = g->nsom;
  int msav, m = g->narc;
  graphe * apm;    /* pour le resultat */
  int *A;          /* tableau pour ranger les index des arcs */
  int i, j, t, q;
  msav = m;

  if (!Connexe(g, g_1)) 
  {   fprintf(stderr, "%s: graphe d'origine non connexe\n", F_NAME);
      exit(0);
  }  

  /* tri des arcs par poids croissant */
  A = (int *)malloc(m * sizeof(int)); /* allocation index */
  if (A == NULL)
  {   fprintf(stderr, "%s: malloc failed\n", F_NAME);
      exit(0);
  }  
  for (i = 0; i < m; i++) A[i] = i; /* indexation initiale */
  TriRapideStochastique(A, g->v_arcs, 0, m-1);

  /* Boucle sur les arcs pris par ordre croissant:
       on enleve un arc, si cela deconnecte le graphe on le remet.
     Arret lorsque m = n-1. */
  i = 0;
  while (m > n-1)
  {
    t = g->tete[A[i]]; q = g->queue[A[i]]; 
    RetireArc(g, t, q);
    RetireArc(g_1, q, t);
    if (Connexe(g, g_1)) 
      m--; 
    else 
    {
      AjouteArc(g, t, q);
      AjouteArc(g_1, q, t);
    }
    i++;
  } // while (m > n-1)

  /* construit le graphe resultat */
  apm = InitGraphe(n, m);
  for (i = 0, j = 0; i < msav; i++)
  {
    t = g->tete[i]; q = g->queue[i]; 
    if (EstSuccesseur(g, t, q)) 
    {
      AjouteArc(apm, t, q);
      apm->tete[j] = t; 
      apm->queue[j] = q; 
      apm->v_arcs[j] = g->v_arcs[i]; 
      j++;
    }
  } // for i

  if (g->x)
    for (i = 0; i < n; i++) { apm->x[i] = g->x[i];  apm->y[i] = g->y[i]; }

  free(A);
  return apm;
} /* Kruskal2() */

/* ====================================================================== */
/*! \fn void DepthTree(graphe * g, graphe *g_1, int a, TYP_VARC *depth, int *farthest)
    \param g (entre) : un graphe (arbre), symtrique ou non
    \param a (entre) : un sommet du graphe g
    \param depth (sortie) : la profondeur de l'arbre de racine a
    \param farthest (sortie) : un sommet le plus loign de a 
    \brief calcule la profondeur de l'arbre de racine a
*/
void DepthTree(graphe * g, int a, TYP_VARC *depth, int *farthest)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "DepthTree"
{
  Lifo * T;   /* liste temporaire geree en pile (Last In, First Out) */
  int i, f, n, s;
  pcell p;
  TYP_VARC* dist;
  TYP_VARC d;

  n = g->nsom;
  dist = (TYP_VARC *)malloc(n * sizeof(TYP_VARC));
  T = CreeLifoVide(n);
  for (i = 0; i < n; i++) dist[i] = -1;

  d = dist[a] = 0;
  LifoPush(T, a);  
  while (!LifoVide(T))
  {
    i = LifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (dist[s] == -1) 
      {
        dist[s] = dist[i] + p->v_arc;
	if (dist[s] > d) { d = dist[s]; f = s; }
        LifoPush(T, s);
      }
    } // for p
  } // while (!LifoVide(T))

  free(dist);
  LifoTermine(T);
  *depth = d;
  *farthest = f;
} /* DepthTree() */

/* ====================================================================== */
/*! \fn void MaxDiameterTree(graphe * g)
    \param g (entre) : un graphe (arbre) symtrique
    \brief calcule un diametre maximal
*/
graphe * MaxDiameterTree(graphe * g)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "MaxDiameterTree"
{
  int n = g->nsom;
  int i, imd;
  TYP_VARC * diam, d, maxdiam = 0;
  int *opp_som, f;
  graphe *pcc;

  diam = (TYP_VARC *)malloc(n * sizeof(TYP_VARC *));
  opp_som = (int *)malloc(n * sizeof(int *));
  for (i = 0; i < n; i++) diam[i] = -1;
  
  for (i = 0; i < n; i++)
    if (Degre(g, i) == 1)
    {
      DepthTree(g, i, &d, &f);
      diam[i] = d;
      opp_som[i] = f;
      if (d > maxdiam) { maxdiam = d; imd = i; }
    }
  Lee(g, imd);
  pcc = PCCna(g, imd, opp_som[imd]);
  return pcc;
} /* MaxDiameterTree() */

/* ====================================================================== */
/*! \fn int LCA(graphe * g, int i, int j)
    \param g (entre) : un graphe (arbre orient des feuilles vers la racine)
    \param i (entre) : un sommet de \b g
    \param j (entre) : un sommet de \b g
    \brief retourne le plus proche anctre commun de i et j (least common ancestor)
*/
int LCA(graphe * g, int i, int j)
/* ====================================================================== */
/* algorithme linaire, efficace pour quelques requtes mais 
   pour un algo en temps constant avec pre-processing linaire voir:
   author = 	 {M.A. Bender and M. Farach-Colton},
   title = 	 {The {LCA} problem revisited},
   booktitle = 	 {procs. Latin American Theoretical Informatics, LNCS, Springer Verlag},
   volume = {1776},
   pages = 	 {88--94},
   year = 	 {2000}   
 */
{
#undef F_NAME
#define F_NAME "LCA"  
#define CHECK_TREE
  int k = -1, n = g->nsom;
  pcell p;
  boolean *M = EnsembleVide(n);

  M[i] = TRUE; // met i et les anctres de i dans l'ensemble M
  p = g->gamma[i]; 
  while (p != NULL)
  {
#ifdef CHECK_TREE
    if (p->next != NULL)
    {
      fprintf(stderr,"%s: non-tree structure\n", F_NAME);
      exit(0);
    }
#endif
    i = p->som;
    M[i] = TRUE;
    p = g->gamma[i]; 
  } 

  if (M[j]) k = j; // recherche un element de M parmi les anctres de j
  else
  {
    p = g->gamma[j]; 
    while (p != NULL) 
    {
#ifdef CHECK_TREE
      if (p->next != NULL)
      {
	fprintf(stderr,"%s: non-tree structure\n", F_NAME);
	exit(0);
      }
#endif
      j = p->som;
      if (M[j]) { k = j; break; }
      p = g->gamma[j];
    } 
  }

#ifdef CHECK_TREE
  if (k == -1)
  {
    fprintf(stderr,"%s: no LCA found\n", F_NAME);
    exit(0);
  }
#endif

  free(M);
  return k;
} /* LCA() */

/* ====================================================================== */
/*! \fn graphe * RootTree(graphe * g, graphe * g_1, int i)
    \param g (entre) : un graphe
    \param g_1 (entre) : le graphe symtrique du graphe g
    \param i (entre) : un sommet de \b g
    \brief retourne un arbre orient (arboresence) de racine i
    dont les arcs correspondent aux artes du graphe non-orient reprsent par g et g_1.
    Les distances (nombre d'arcs) de la racine aux sommets sont de plus stocks dans les champs v_sommets.
*/
graphe * RootTree(graphe * g, graphe * g_1, int i)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "RootTree"
{
  int k, s, n;
  pcell p;
  Fifo * T;
  graphe * tree;

  if ((g == NULL) || (g_1 == NULL))
  {
    fprintf(stderr,"%s: input graph not available\n", F_NAME);
    exit(0);
  }
  n = g->nsom;

  T = CreeFifoVide(n);
  if (T == NULL)
  {
    fprintf(stderr,"%s: CreeFifoVide failed\n", F_NAME);
    exit(0);
  }

  tree = InitGraphe(n, n-1);
  if (tree == NULL)
  {
    fprintf(stderr,"%s: InitGraphe failed\n", F_NAME);
    exit(0);
  }

  for (k = 0; k < n; k++) tree->v_sommets[k] = -1;
  tree->v_sommets[i] = 0;
  FifoPush(T, i);

  while (!FifoVide(T))
  {
    i = FifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (tree->v_sommets[s] == -1) 
      {
	AjouteArcValue(tree, i, s, p->v_arc);
	tree->v_sommets[s] = tree->v_sommets[i] + 1;
	FifoPush(T, s);
      }
    } // for p
    for (p = g_1->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s prdcesseur de i */
      s = p->som;
      if (tree->v_sommets[s] == -1) 
      {
	AjouteArcValue(tree, i, s, p->v_arc);
	tree->v_sommets[s] = tree->v_sommets[i] + 1;
	FifoPush(T, s);
      }
    } // for p
  } // while (!FifoVide(T))

  FifoTermine(T);
  return tree;
} /* RootTree() */

/* ====================================================================== */
/*! \fn void DistRoot(graphe * g, int r)
    \param g (entre) : un graphe (arbre orient ou arborescence)
    \param r (entre) : racine de \b g
    \brief calcule, pour chaque sommet x de g, la longueur (somme des poids d'arcs) d'un plus court
           chemin de r vers x. Cette longueur est stocke dans le champ
           \b v_sommets de \b g
*/
void DistRoot(graphe * g, int r)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "DistRoot"
{
  int k, s, n = g->nsom;
  pcell p;
  Fifo * T = CreeFifoVide(n);

  for (k = 0; k < n; k++) g->v_sommets[k] = -1;
  g->v_sommets[r] = 0;
  FifoPush(T, r);

  while (!FifoVide(T))
  {
    r = FifoPop(T);
    for (p = g->gamma[r]; p != NULL; p = p->next)
    { /* pour tout s successeur de r */
      s = p->som;
      if (g->v_sommets[s] == -1) 
      {
        g->v_sommets[s] = g->v_sommets[r] + p->v_arc;
        FifoPush(T, s);
      }
    } // for p
  } // while (!FifoVide(T))

  FifoTermine(T);
} /* DistRoot() */

/* ====================================================================== */
/* ====================================================================== */
/* PLUS COURTS CHEMINS */
/* ====================================================================== */
/* ====================================================================== */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2, 3 */
/* ====================================================================== */
/* \fn void Dijkstra1(graphe * g, int i)
    \param g (entre) : un graphe pondr. La pondration de chaque arc doit 
                        se trouver dans le champ \b v_arc de la structure \b cell
    \param i (entre) : un sommet de \b g
    \brief calcule, pour chaque sommet x de g, la longueur d'un plus court
           chemin de i vers x. Cette longueur est stocke dans le champ
           \b v_sommets de \b g
*/
void Dijkstra1(graphe * g, int i)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Dijkstra1"
{
  int n = g->nsom;
  boolean * S = EnsembleVide(n);
  int k, x, y;
  pcell p;
  TYP_VSOM vmin;

  S[i] = TRUE;
  for (k = 0; k < n; k++) g->v_sommets[k] = L_INFINITY;
  g->v_sommets[i] = 0;
  k = 1;
  x = i;
  while ((k < n) && (g->v_sommets[x] < L_INFINITY))
  {
    for (p = g->gamma[x]; p != NULL; p = p->next)
    { /* pour tout y successeur de x */
      y = p->som;
      if (!S[y])
      {
        g->v_sommets[y] = min((g->v_sommets[y]),(g->v_sommets[x]+p->v_arc));
      } // if (!S[y])
    } // for p
    // extraire un sommet x hors de S de valeur g->v_sommets minimale
    vmin = L_INFINITY;
    for (y = 0; y < n; y++)
      if (!S[y]) 
        if (g->v_sommets[y] <= vmin) { x = y; vmin = g->v_sommets[y]; }
    k++; 
    S[x] = TRUE;
  } // while ((k < n) && (g->v_sommets[x] < L_INFINITY))

} /* Dijkstra1() */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2, 3 */
/* ====================================================================== */
/*! \fn void Dijkstra(graphe * g, int i)
    \param g (entre) : un graphe pondr. La pondration de chaque arc doit 
                        se trouver dans le champ \b v_arc de la structure \b cell
    \param i (entre) : un sommet de \b g
    \brief calcule, pour chaque sommet x de g, la longueur d'un plus court
           chemin de i vers x. Cette longueur est stocke dans le champ
           \b v_sommets de \b g
*/
void Dijkstra(graphe * g, int i)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Dijkstra"
{
  int n = g->nsom;
  boolean * S = EnsembleVide(n);
  int k, x, y;
  pcell p;
  TYP_VSOM vmin;
  int *T = ListeVide(n);
  int tx, ty, tt = 0;

  S[i] = TRUE;
  for (k = 0; k < n; k++) g->v_sommets[k] = L_INFINITY;
  g->v_sommets[i] = 0;
  x = i;
  do
  {
    for (p = g->gamma[x]; p != NULL; p = p->next)
    { /* pour tout y successeur de x */
      y = p->som;
      if (!S[y])
      {
        g->v_sommets[y] = min((g->v_sommets[y]),(g->v_sommets[x]+p->v_arc));
        T[tt] = y; // insere y dans T
        tt++;
      } // if (!S[y])
    } // for p

    // extraire un sommet x de T de valeur g->v_sommets minimale
    tx = 0; 
    x = T[0];
    vmin = g->v_sommets[x];
    for (ty = 1; ty < tt; ty++)
    {
      y = T[ty];
      if (g->v_sommets[y] <= vmin) { x = y; tx = ty; vmin = g->v_sommets[y]; }
    }
    S[x] = TRUE;
    for (; tx < tt-1; tx++) T[tx] = T[tx+1]; // retire x de T
    tt--;
  } while (tt > 0);
  free(T);

} /* Dijkstra() */

/* ====================================================================== */
/*! \fn void Lee(graphe * g, int i)
    \param g (entre) : un graphe
    \param i (entre) : un sommet de \b g
    \brief calcule, pour chaque sommet x de g, la longueur (nombre d'arcs) d'un plus court
           chemin de i vers x. Cette longueur est stocke dans le champ
           \b v_sommets de \b g
*/
void Lee(graphe * g, int i)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "Lee"
{
  int k, s, n = g->nsom;
  pcell p;
  Fifo * T = CreeFifoVide(n);

  for (k = 0; k < n; k++) g->v_sommets[k] = -1;
  g->v_sommets[i] = 0;
  FifoPush(T, i);

  while (!FifoVide(T))
  {
    i = FifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (g->v_sommets[s] == -1) 
      {
        g->v_sommets[s] = g->v_sommets[i] + 1;
        FifoPush(T, s);
      }
    } // for p
  } // while (!FifoVide(T))

  FifoTermine(T);
} /* Lee() */

/* ====================================================================== */
/*! \fn void LeeNO(graphe * g, graphe * g_1, int i)
    \param g (entre) : un graphe
    \param g_1 (entre) : le graphe symtrique du graphe g
    \param i (entre) : un sommet de \b g
    \brief calcule, pour chaque sommet x de g, la longueur (nombre d'artes) 
    d'un plus court chemin de i vers x dans le graphe non orient reprsent 
    par (g, g_1). Cette longueur est stocke dans le champ \b v_sommets de \b g
*/
void LeeNO(graphe * g, graphe * g_1, int i)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "LeeNO"
{
  int k, s, n = g->nsom;
  pcell p;
  Fifo * T = CreeFifoVide(n);

  for (k = 0; k < n; k++) g->v_sommets[k] = -1;
  g->v_sommets[i] = 0;
  FifoPush(T, i);

  while (!FifoVide(T))
  {
    i = FifoPop(T);
    for (p = g->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s successeur de i */
      s = p->som;
      if (g->v_sommets[s] == -1) 
      {
        g->v_sommets[s] = g->v_sommets[i] + 1;
        FifoPush(T, s);
      }
    } // for p
    for (p = g_1->gamma[i]; p != NULL; p = p->next)
    { /* pour tout s prdcesseur de i */
      s = p->som;
      if (g->v_sommets[s] == -1) // sic: g
      {
        g->v_sommets[s] = g->v_sommets[i] + 1;  // sic: g
        FifoPush(T, s);
      }
    } // for p
  } // while (!FifoVide(T))

  FifoTermine(T);
} /* LeeNO() */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2, 3 */
/* ====================================================================== */
/*! \fn graphe * PCC(graphe * g, int d, int a)
    \param g (entre) : un graphe pondr, reprsent par son application successeurs,
            et dont les sommets ont t valus par la distance au sommet \b d
    \param d (entre) : un sommet (dpart)
    \param a (entre) : un sommet (arrive)
    \return un plus court chemin de \b d vers \b a dans \b g , reprsent par un graphe
    \brief retourne un plus court chemin de \b d vers \b a dans \b g
*/
graphe * PCC(graphe * g, int d, int a)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "PCC"
{
  int n = g->nsom;
  int m = g->narc;
  graphe * pcc = InitGraphe(n, n-1); /* pour le resultat */
  graphe * g_1 = Symetrique(g);
  int i, y, x = a;
  pcell p;

  while (x != d)
  {
    for (p = g_1->gamma[x]; p != NULL; p = p->next)
    { /* pour tout y predecesseur de x */
      y = p->som;
      if ((g->v_sommets[x]-g->v_sommets[y]) == p->v_arc) 
      {
        AjouteArcValue(pcc, y, x, p->v_arc);
        x = y;
        break;
      }
    } // for p
    if (p == NULL)
    {
      printf("WARNING: pas de pcc trouve de %d vers %d\n", d, a);
      TermineGraphe(g_1);
      return pcc;
    }
  }
  TermineGraphe(g_1);
  for (i = 0; i < n; i++) { pcc->x[i] = g->x[i]; pcc->y[i] = g->y[i]; }
  return pcc;
} /* PCC() */

/* ====================================================================== */
/* A RETIRER POUR LES TP 1, 2, 3 */
/* ====================================================================== */
/*! \fn graphe * PCCna(graphe * g, int d, int a)
    \param g (entre) : un graphe, reprsent par son application successeurs,
            et dont les sommets ont t valus par la distance (nombre d'arcs) au sommet \b d
    \param d (entre) : un sommet (dpart)
    \param a (entre) : un sommet (arrive)
    \return un plus court chemin (au sens du nombre d'arcs) de \b d vers \b a dans \b g , 
    reprsent par un graphe
    \brief retourne un plus court chemin de \b d vers \b a dans \b g
*/
graphe * PCCna(graphe * g, int d, int a)
/* ====================================================================== */
#undef F_NAME
#define F_NAME "PCCna"
{
  int n = g->nsom;
  int m = g->narc;
  graphe * pcc = InitGraphe(n, n-1); /* pour le resultat */
  graphe * g_1 = Symetrique(g);
  int i, y, x = a;
  pcell p;

  while (x != d)
  {
    for (p = g_1->gamma[x]; p != NULL; p = p->next)
    { /* pour tout y predecesseur de x */
      y = p->som;
      if ((g->v_sommets[x]-g->v_sommets[y]) == 1) 
      {
        AjouteArcValue(pcc, y, x, p->v_arc);
        x = y;
        break;
      }
    } // for p
    if (p == NULL)
    {
      printf("WARNING: pas de pcc trouve \n");
      TermineGraphe(g_1);
      return pcc;
    }
  }
  TermineGraphe(g_1);
  for (i = 0; i < n; i++) { pcc->x[i] = g->x[i]; pcc->y[i] = g->y[i]; }
  return pcc;
} /* PCCna() */

/* ====================================================================== */
/* ====================================================================== */
/* PROGRAMME DE TEST */
/* ====================================================================== */
/* ====================================================================== */

#ifdef TEST
int main(int argc, char **argv)
{
  graphe * g, *g_1, *a;
  int s1, s2, na, ns;
  boolean *Cs;

  if (argc != 3)
  {
    fprintf(stderr, "usage: %s <nombre sommets> <nombre arcs>\n", argv[0]);
    exit(0);
  }

  ns = atoi(argv[1]);
  na = atoi(argv[2]);
  
  g = GrapheAleatoire(ns, na);
  AfficheSuccesseurs(g);
  AfficheArcs(g);
  g_1 = Symetrique(g);
  AfficheSuccesseurs(g_1);
  AfficheArcs(g_1);

  if (Connexe(g, g_1))
  {
    printf("graphe connexe\n");
    a = Kruskal2(g, g_1);
    AfficheSuccesseurs(a);
    AfficheArcs(a);
    TermineGraphe(a);
  }
  else printf("graphe NON connexe\n");

  do
  {
    printf("entrer un sommet : ");
    scanf("%d", &s1);
    if (s1 >= 0) 
    {
      Cs = CompFortConnexe(g, g_1, s1);
      AfficheEnsemble(Cs, g->nsom);
      free(Cs);
    }
  } while (s1 >= 0);

  do
  {
    printf("entrer un sommet : ");
    scanf("%d", &s1);
    if (s1 >= 0) 
    {
      Cs = CompConnexe(g, g_1, s1);
      AfficheEnsemble(Cs, g->nsom);
      free(Cs);
    }
  } while (s1 >= 0);

  TermineGraphe(g);
  TermineGraphe(g_1);

} /* main() */
#endif
